"""Main components of sparkplug standard"""
import logging

from aiospb.nodes.messages import MqttNodeCarrier

from .scanning import Scanner
from .states import Disconnected, NodeContext
from .stores import HistoricalStore

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MetricsConnectionError(Exception):
    ...


class EdgeNode:
    """Gateway connected to mqtt and to one or more hardware devices"""

    def __init__(
        self,
        name: str,
        mqtt_carrier: MqttNodeCarrier,
        scanner: Scanner,
        store: HistoricalStore,
        primary_hostname: str = "",
    ):
        self._name = name
        self._carrier = mqtt_carrier
        self._scanner = scanner
        self._store = store
        self._primary_hostname = primary_hostname
        self._context = NodeContext(
            self._carrier,
            self._scanner,
            self._store,
            self._primary_hostname,
        )
        self._context.state = Disconnected(self._context)

    @property
    def name(self) -> str:
        return self._name

    @property
    def state(self) -> str | None:
        if self._context is None:
            return
        return self._context.state.__class__.__name__.lower()

    async def establish_session(self):
        """Establish a session for a supplied metrics net"""
        if self._context.state is None:  # pragma: no cover
            return

        await self._context.state.establish_session()

    async def terminate_session(self):
        """Terminate cleanly the session and"""
        if self._context.state is None:  # pragma: no cover
            return

        await self._context.state.terminate_session()



import asyncio
import logging
from typing import Callable, Coroutine, Sequence

from aiospb import Clock, UtcClock
from aiospb.data import DataType, Metric, WriteRequest
from aiospb.mqtt import (
    HostPayload,
    MqttClient,
    MqttError,
    NodePayload,
    SpbMessage,
    Topic,
    Will,
)

from .commands import CommandRequest, CommandResponse, NodeCommander, WritingResolution
from .sorting import NodeMessageSorter

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class _Observer:
    def __init__(self, callback: Callable, filter: str = ""):
        self.callback = callback
        self._filter = filter

    async def notify(self, message: SpbMessage):
        if not self._filter or self._filter in message.topic.component_name:
            await self.callback(message)


class HostBridge:
    """Easy interface to host applications"""

    def __init__(
        self,
        hostname: str,
        mqtt_client: MqttClient,
        groups: str = "",
        reorder_timeout: float = 2,
        clock: Clock | None = None,
    ):
        self._hostname = hostname
        self._groups = groups.split(",")
        self._mqtt_client = mqtt_client
        self._reorder_timeout = reorder_timeout
        self._clock = clock if clock else UtcClock()

        self._observers = []
        self._is_online = False
        self._recieving = None
        self._sorters = {}
        self._commanders = {}

    @property
    def hostname(self) -> str:
        """Name of the host application"""
        return self._hostname

    @property
    def is_online(self) -> bool:
        return self._is_online

    async def establish_session(self, groups: list[str] | None = None):
        """Init session to listen edge nodes"""

        try:
            await self._connect()
            groups = groups or ["+"]
            await self._subscribe(groups)
            self._recieving = asyncio.create_task(self._recieve_node_messages())
            await self._send_birth_certificate()
        except MqttError as e:
            logger.error("Error with the MQTT Server when establishing session")
            logger.exception(e)
            self._is_online = False
            raise e

        logger.info(f'Host application "{self._hostname}" has established session')
        self._is_online = True

    async def terminate_session(self):
        """Close cleanly a session"""
        if self._recieving is not None:
            self._recieving.cancel()

        try:
            await self._send_death_certificate()
        except MqttError as e:
            logger.error("MQTT connection broken while sending host death certificate")
            logger.exception(e)

        logger.info(f'Host application "{self._hostname}" has terminated session')
        self._is_online = False

    def observe_nodes(
        self,
        callback: Callable[[SpbMessage], Coroutine[None, None, None]],
        node_filter: str = "",
    ) -> None:
        """Add one callable observer when it rece"""

        self._observers.append(_Observer(callback, node_filter))

    async def _recieve_node_messages(self):
        while True:
            try:
                message: SpbMessage = await self._mqtt_client.deliver_message()
                if not message.is_from_node():
                    continue

                node_name = message.topic.component_name
                if node_name not in self._sorters:
                    self._commanders[node_name] = NodeCommander(
                        node_name, self._mqtt_client, self._clock
                    )
                    self._sorters[node_name] = NodeMessageSorter(
                        node_name, self._clock, self._reorder_timeout
                    )

                sorter = self._sorters[node_name]
                sorter.register_message(message)
                try:
                    for message in sorter.nexts():
                        self._commanders[node_name].handle_node_message(message)
                        await self._notify_message_to_observers(message)
                except TimeoutError:
                    await self._send_rebirth_command(node_name)
                    logger.info(
                        f"Sent rebirth command to {node_name} by lost of message order"
                    )
            except MqttError as e:
                logger.error(f"Error when recieving node messages")
                logger.exception(e)
                self._is_online = False
                return
            except Exception as e:
                logger.error("Unmanaged exception when recieving node messages")
                logger.exception(e)
                logger.critical("The host is stopping to recieve messages")
                self._is_online = False
                return

    async def write_metrics(
        self, node_name: str, write_requests: list[WriteRequest], timeout: float = 10.0
    ) -> CommandResponse:
        """Request changes to metrics"""
        return await self._commanders[node_name].execute(
            CommandRequest(write_requests, timeout)
        )

    async def _notify_message_to_observers(self, message):
        results = await asyncio.gather(
            *[observer.notify(message) for observer in self._observers],
            return_exceptions=True,
        )
        for result, observer in zip(results, self._observers):
            if type(result) is Exception:
                logger.warning(f"Observer {observer} raised exception when notified")
                logger.exception(result)

    async def _connect(self):
        await self._mqtt_client.connect(
            self._hostname,
            will=Will(
                SpbMessage(
                    Topic.from_component(self._hostname, "STATE"),
                    HostPayload(self._clock.timestamp(), False),
                ),
                qos=1,
                retain=True,
            ),
        )

    async def _subscribe(self, groups):
        for group in groups:
            await self._mqtt_client.subscribe(f"spBv1.0/{group}/+/+", qos=1)

    async def _send_birth_certificate(self):
        await self._mqtt_client.publish(
            SpbMessage(
                Topic.from_component(self._hostname, "STATE"),
                HostPayload(self._clock.timestamp(), True),
            ),
            qos=1,
            retain=True,
        )

    async def _send_rebirth_command(self, node_name):
        ts = self._clock.timestamp()
        message = SpbMessage(
            Topic.from_component(node_name, "NCMD"),
            NodePayload(
                ts,
                metrics=[
                    Metric(
                        ts,
                        True,
                        DataType.Boolean,
                        name="Node Control/Rebirth",
                    )
                ],
            ),
        )
        await self._mqtt_client.publish(
            message,
            qos=0,
            retain=False,
        )

    async def _send_death_certificate(self):
        await self._mqtt_client.publish(
            SpbMessage(
                Topic.from_component(self._hostname, "STATE"),
                HostPayload(self._clock.timestamp(), False),
            ),
            qos=1,
            retain=True,
        )
        await self._mqtt_client.disconnect()



"""A MQTT Server can not warrant that host application would recieve the node messages
in the same order is sent. If not, standards rules a rebirth"""
import asyncio
import logging

from aiospb import Clock, UtcClock
from aiospb.data import DataType, Metric, WriteRequest
from aiospb.mqtt import (
    HostPayload,
    MqttClient,
    MqttError,
    NodePayload,
    SpbMessage,
    Topic,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class NodeMessageSorter:
    def __init__(
        self,
        node_name: str,
        clock: Clock,
        reorder_time: float = 2.0,  # in s
    ):
        self._node_name = node_name
        self._clock = clock
        self._reorder_time = reorder_time

        self._seq = -1
        self._bd_seq = -1
        self._messages = {}
        self._timeout_ts = 0

    def register_message(self, message: SpbMessage):
        if type(message.payload) is not NodePayload:
            return

        if message.is_a("NCMD"):
            return

        if message.is_a("NDATA") and self._seq == -1:
            logger.warning(f"Lost of data by no birth at {self._node_name}")
            return

        if message.is_a("NDEATH"):
            if message.payload.metrics[0].value == self._bd_seq:
                self._messages[None] = message
            return

        if message.is_a("NBIRTH"):
            self._messages.clear()

            bd_seq = None
            for metric in message.payload.metrics:
                if metric.name == "bdSeq":
                    bd_seq = metric.value

            if bd_seq is None:
                raise ValueError(f"Birth from {self._node_name} has not bdSeq!!")

            if bd_seq != self._bd_seq:
                self._bd_seq = bd_seq
                self._seq = message.payload.seq
                self._messages.clear()

        self._messages[message.payload.seq] = message

    def nexts(self) -> list[SpbMessage]:
        output = []
        while self._messages:
            if self._seq in self._messages:
                output.append(self._messages.pop(self._seq))
                self._seq = 0 if self._seq == 255 else self._seq + 1
                self._timeout_ts = self._clock.timestamp() + self._reorder_time * 1000
            elif None in self._messages:
                message = self._messages.pop(None)
                output.append(message)
                self._messages.clear()
                self._seq = -1
            else:
                if self._clock.timestamp() > self._timeout_ts:
                    self._messages.clear()
                    logger.warning(
                        f"There are {len(self._messages)} messages disordered too much time in {self._node_name}"
                    )
                    raise TimeoutError()
                break
        return output



import json

from .. import HostPayload, MessageEncoder, NodePayload, Payload


class JsonEncoder(MessageEncoder):
    """Encoder of messages for testing purposes"""

    def encode(self, payload: Payload) -> bytes:
        """Convert a payload dict to a string for publishing"""

        return json.dumps(payload.to_dict(), sort_keys=True).encode("utf-8")

    def decode(self, payload: bytes) -> Payload:
        """Convert payload to a payload dict"""
        payload_map = json.loads(payload.decode("utf-8"))
        if "online" in payload_map:
            return HostPayload.from_dict(payload_map)
        else:
            return NodePayload.from_dict(payload_map)



import abc
import asyncio
import logging
from dataclasses import dataclass, field
from typing import Self

from .. import Clock, UtcClock
from ..data import DataType, Metric, PropertySet, PropertyValue, Quality, ValueType

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class MetricCore:
    """Metric data, stable for a session"""

    name: str
    datatype: DataType
    properties: PropertySet = field(default_factory=PropertySet)
    alias: int = 0
    is_transient: bool = False

    def create_metric(
        self, value: ValueType, timestamp: int, quality: Quality | None = None
    ) -> Metric:
        """Create metric from the metric core"""
        props = self.properties
        if quality is not None:
            props_data = self.properties.as_dict()
            props_data["quality"] = {"value": quality.value, "dataType": "Int32"}
            props = PropertySet.from_dict(props_data)

        return Metric(
            timestamp,
            value,
            self.datatype,
            self.alias,
            self.name,
            props,
            self.is_transient,
        )

    # @classmethod
    # def from_dict(cls, dump: dict[str, Any]) -> Self:
    #     datatype = (
    #         DataType(dump["dataType"])
    #         if type(dump["dataType"]) is int
    #         else DataType[dump["dataType"]]
    #     )
    #     properties = (
    #         PropertySet.from_dict(dump["properties"])
    #         if "properties" in dump
    #         else PropertySet()
    #     )
    #     return cls(
    #         dump["name"],
    #         datatype,
    #         properties,
    #         dump.get("alias", 0),
    #         dump.get("is_transient", False),
    #     )


class MetricsNet(abc.ABC):
    # @abc.abstractmethod
    # async def smap(self) -> dict[str, int]:
    #     """List all name of metrics available and their aliases, 0 it does not have"""

    @abc.abstractmethod
    async def get_metric_cores(self, filter: str = "") -> list[MetricCore]:
        """List available metrics in the net with all its properties"""

    @abc.abstractmethod
    async def read_value(self, metric_name: str = "", alias: int = 0) -> ValueType:
        """Read a value to a metric"""

    @abc.abstractmethod
    async def write_value(
        self, value: ValueType, metric_name: str = "", alias: int = 0
    ):
        """Write a value to a metric"""


@dataclass
class _Sample:
    timestamp: int
    value: ValueType
    quality: Quality

    @classmethod
    async def read(
        cls,
        metrics_net: MetricsNet,
        clock: Clock,
        metric_name: str = "",
        alias: int = 0,
    ) -> Self:
        if not alias and not metric_name:
            raise ValueError("Metric Name and alias can not be null at the same time")
        try:
            value = await metrics_net.read_value(metric_name)
            return cls(clock.timestamp(), value, Quality.GOOD)
        except asyncio.CancelledError:
            return cls(clock.timestamp(), None, Quality.STALE)
        except Exception as e:
            logger.error(f"Exception when reading metric {metric_name}:{alias}")
            logger.exception(e)
            return cls(clock.timestamp(), None, Quality.BAD)


@dataclass
class _Scan:
    scan_rate: int
    datatype: DataType
    quality: Quality
    value: ValueType = None
    alias: int = 0
    name: str = ""

    def update(self, sample: _Sample) -> Metric | None:
        if sample.quality == self.quality and sample.value == self.value:
            return

        if self.quality != Quality.GOOD and sample.quality == self.quality:
            return

        props = (
            PropertySet()
            if self.quality == sample.quality
            else PropertySet(
                ("quality",), (PropertyValue(sample.quality.value, DataType.Int32),)
            )
        )
        self.quality = sample.quality

        value = sample.value if sample.quality == Quality.GOOD else self.value
        self.value = value

        name = "" if self.alias else self.name
        return Metric(sample.timestamp, value, self.datatype, self.alias, name, props)

    def plan_next_scan(self, schedules: dict[int, list[Self]], clock: Clock):
        timestamp = ((clock.timestamp() // self.scan_rate) + 1) * self.scan_rate
        if timestamp not in schedules:
            schedules[timestamp] = []

        schedules[timestamp].append(self)


class Scanner:
    def __init__(
        self,
        metrics_net: MetricsNet,
        scan_rate: int = 60000,
        clock: Clock | None = None,
    ):
        self._net = metrics_net
        self._clock = clock or UtcClock()
        self._scan_rate = scan_rate

        self._queue = asyncio.Queue()
        self._schedules = {}  # scans ordered by timestamp to be triggered
        self._tasks = {}  # scannings to be done in the future

    @property
    def scan_rate(self) -> float:
        return self._scan_rate

    async def deliver_changes(self, timeout: float = 0.0) -> list[Metric]:
        """Return last changes"""
        if timeout:
            return await asyncio.wait_for(self._queue.get(), timeout=timeout)
        return await self._queue.get()

    async def _read_births(self) -> list[Metric]:
        cores = await self._net.get_metric_cores()

        readings = [
            asyncio.create_task(_Sample.read(self._net, self._clock, core.name))
            for core in cores
        ]
        _, pending = await asyncio.wait(readings, timeout=self._scan_rate / 1000)
        if pending:
            logger.warning(f"{len(pending)} metrics are in STALE quality when birth")
            for core, reading in zip(cores, readings):
                if not reading.done():
                    logger.warning(f"Metric {core.name}:{core.alias} is STALE")
        for reading in pending:
            reading.cancel()
        await asyncio.gather(*pending, return_exceptions=True)

        births = []
        for core, reading in zip(cores, readings):
            sample = reading.result()
            births.append(
                core.create_metric(sample.value, sample.timestamp, sample.quality)
            )

        return births

    async def start(self) -> list[Metric]:
        ts = self._clock.timestamp()
        node_controls = [
            Metric(ts, False, DataType.Boolean, name="Node Control/Reboot"),
            Metric(ts, False, DataType.Boolean, name="Node Control/Rebirth"),
            Metric(
                ts,
                int(self._scan_rate * 1000),
                DataType.Int64,
                name="Node Control/Scan Rate",
            ),
        ]

        births = await self._read_births()
        scans = []
        for birth in births:
            if "Properties/" in birth.name:  # Properties will not be scanned
                continue

            scan_rate = self._scan_rate
            if "scan_rate" in birth.properties:
                scan_rate = birth.properties["scan_rate"].value
                if type(scan_rate) is not int:
                    raise ValueError(f"Scan rate can not be of type {type(scan_rate)}")

                if scan_rate == 0:  # No scan this metric
                    continue

            scans.append(
                _Scan(
                    scan_rate,
                    birth.datatype,
                    Quality(birth.properties["quality"].value),
                    birth.value,
                    alias=birth.alias,
                    name=birth.name,
                )
            )

        self._plan_scans(scans)

        return node_controls + births

    async def _scan_schedule(self, timestamp: int):
        wait_time = (timestamp - self._clock.timestamp()) / 1000
        await self._clock.asleep(wait_time if wait_time > 0 else 0)

        readings = await self._read_metrics(self._schedules[timestamp])
        self._report_changes(readings)
        self._plan_scans(self._schedules.pop(timestamp, []))
        if timestamp in self._tasks:
            self._tasks.pop(timestamp)

    async def _read_metrics(self, scans: list[_Scan]) -> list:
        readings = [
            asyncio.create_task(
                _Sample.read(self._net, self._clock, scan.name, scan.alias)
            )
            for scan in scans
        ]
        logger.debug(f"Scanning {len(readings)} metrics")

        timeout = self._scan_rate / 2000  #  half of scan rate
        try:
            _, pending = await asyncio.wait(
                readings,
                timeout=timeout,
            )
        except asyncio.CancelledError as e:
            for reading in readings:
                if not reading.done():
                    reading.cancel()
            raise e

        if pending:
            logger.warning(f"{len(pending)} metrics are in STALE quality when scanning")
            for scan, reading in zip(scans, readings):
                if not reading.done():
                    logger.warning(f"Metric {scan.name}:{scan.alias} is STALE")

        for reading in pending:
            reading.cancel()
        await asyncio.gather(*pending)

        changes = []
        for scan, reading in zip(scans, readings):
            metric = scan.update(reading.result())
            if metric is not None:
                changes.append(metric)

        return sorted(changes, key=lambda change: change.timestamp)

    def _report_changes(self, changes):
        if changes:
            self._queue.put_nowait(changes)
            logger.info(f"Reported {len(changes)} changes after scanning")
        else:
            logger.info("No changes to be reported after scanning")

    def _plan_scans(self, scans):
        for scan in scans:
            scan.plan_next_scan(self._schedules, self._clock)

        for timestamp in self._schedules.keys():
            if timestamp not in self._tasks:
                self._tasks[timestamp] = asyncio.create_task(
                    self._scan_schedule(timestamp)
                )

        logger.info(f"Planned to scan {len(scans)} metrics")

    async def stop(self):
        """Stop all next scanning tasks"""
        for task in self._tasks.values():
            task.cancel()
        await asyncio.gather(*list(self._tasks.values()), return_exceptions=True)
        self._tasks.clear()
        self._schedules.clear()

    def _find_scan(self, metric: Metric) -> _Scan | None:
        for scans in self._schedules.values():
            for scan in scans:
                if (metric.alias and metric.alias == scan.alias) or (
                    metric.name and metric.name == scan.metric_name
                ):
                    return scan

    async def execute_command(self, metric: Metric) -> Metric:
        """execute a command, writing metric to the net"""

        if metric.name == "Node Control/Scan Rate":
            if type(metric.value) is not int:
                raise ValueError(f"Metric {metric.name} has not type integer")
            self.scan_rate = metric.value
            return Metric(
                self._clock.timestamp(),
                metric.value,
                DataType.Int32,
                name="Node Control/Scan Rate",
            )

        try:
            await self._net.write_value(
                metric.value, metric_name=metric.name, alias=metric.alias
            )
            sample = _Sample(self._clock.timestamp(), metric.value, Quality.GOOD)
        except asyncio.CancelledError:
            sample = _Sample(self._clock.timestamp(), metric.value, Quality.STALE)
        except Exception as e:
            logger.error(
                f"Writing {metric.value} to metric '{metric.name}:{metric.alias}'"
            )
            logger.exception(e)
            sample = _Sample(self._clock.timestamp(), metric.value, Quality.BAD)

        scan = self._find_scan(metric)
        if scan:
            scan.update(sample)
        return Metric(
            sample.timestamp,
            metric.value,
            metric.datatype,
            metric.alias,
            metric.name,
            properties=PropertySet.from_dict(
                {"quality": {"value": sample.quality.value, "dataType": "Int32"}}
            ),
        )

    @scan_rate.setter
    def scan_rate(self, value: int):
        replanned_scans = []
        for timestamp in list(self._schedules.keys()):
            scans = []
            for scan in self._schedules.pop(timestamp):
                if scan.scan_rate == self._scan_rate:
                    replanned_scans.append(
                        _Scan(
                            value,
                            scan.datatype,
                            scan.quality,
                            scan.value,
                            scan.alias,
                            scan.name,
                        )
                    )
                else:
                    scans.append(scan)
            if scans:
                self._schedules[timestamp] = scans
            else:
                self._tasks.pop(timestamp).cancel()

        self._plan_scans(replanned_scans)
        self._scan_rate = value



import asyncio
import logging
from dataclasses import dataclass
from enum import Enum

from aiospb import Clock
from aiospb.data import DataType, Metric, Quality, ValueType, WriteRequest
from aiospb.mqtt import MqttClient, MqttError, NodePayload, SpbMessage, Topic
from aiospb.mqtt.core import MqttClient, NodePayload, SpbMessage

logger = logging.getLogger(__name__)


@dataclass
class CommandRequest:
    requests: list[WriteRequest]
    writing_timeout: float = 60.0


class WritingResolution(Enum):
    BadWriting = 0
    EdgeIsOffline = 5
    HostIsOffline = 10
    MetricNotFound = 15
    GoodWriting = 192
    StaleWriting = 500
    WritingTimeout = 505


@dataclass
class CommandResponse:
    timestamp: int
    resolutions: list[WritingResolution]


class _CommandTrack:
    def __init__(self, metric_ids: list[int | str], values: list[ValueType]):
        self.metric_ids = metric_ids
        self.values = values
        self.has_finished = asyncio.Event()
        self.resolutions = [WritingResolution.WritingTimeout] * len(values)

    def confirm_change(self, metric: Metric):
        index = None

        if metric.alias:
            try:
                index = self.metric_ids.index(metric.alias)
            except ValueError:
                pass

        if index is None:
            try:
                index = self.metric_ids.index(metric.name)
            except ValueError:
                return

        if self.values[index] != metric.value:
            return

        if "quality" in metric.properties:
            self.resolutions[index] = WritingResolution(metric.properties["quality"])

        if WritingResolution.WritingTimeout not in self.resolutions:
            self.has_finished.set()


class NodeCommander:
    def __init__(self, node_name: str, mqtt_client: MqttClient, clock: Clock):
        self._node_name = node_name
        self._mqtt_client = mqtt_client
        self._clock = clock
        self._aliases = {}
        self._datatypes = {}
        self._tracks = {}
        self._resolutions = {}
        self._is_online = False

    def handle_node_message(self, message: SpbMessage):
        """Update info of open commands"""
        if type(message.payload) is not NodePayload:
            return

        if message.is_a("NBIRTH"):
            self._is_online = True
            self._aliases.clear()
            self._datatypes.clear()
            for metric in message.payload.metrics:
                if "Properties" in metric.name:
                    continue
                if metric.alias:
                    self._aliases[metric.name] = metric.alias
                    self._datatypes[metric.alias] = metric.datatype
                else:
                    self._datatypes[metric.name] = metric.datatype
            return

        if message.is_a("NDATA"):
            if self._datatypes:
                logger.warning(
                    (
                        f"Recieved {len(message.payload.metrics)} without birth"
                        f" from {self._node_name}, ignoring..."
                    )
                )

            for metric in message.payload.metrics:
                for track in self._tracks.values():
                    track.confirm_change(metric)

        if message.is_a("NDEATH"):
            self._is_online = False

    async def execute(self, request: CommandRequest) -> CommandResponse:
        """Execute a command request"""
        n_writes = len(request.requests)
        if not self._is_online:
            return CommandResponse(
                self._clock.timestamp(),
                [WritingResolution.EdgeIsOffline] * n_writes,
            )

        try:
            timestamp = self._clock.timestamp()
            await self._send_command(timestamp, request.requests)
        except MqttError:
            return CommandResponse(
                self._clock.timestamp(), [WritingResolution.HostIsOffline] * n_writes
            )

        try:
            await asyncio.wait_for(
                self._tracks[timestamp].has_finished.wait(),
                timeout=request.writing_timeout - timestamp,
            )
        except TimeoutError:
            pass

        return CommandResponse(
            self._clock.timestamp(), self._tracks.pop(timestamp).resolutions
        )

    async def _send_command(self, timestamp: int, requests: list[WriteRequest]):
        metrics = []
        metric_ids = []
        values = []

        not_found_metrics = []
        for index, request in enumerate(requests):
            name = ""
            alias = 0

            if request.metric_name and request.metric_name in self._aliases:
                alias = self._aliases[request.metric_name]

            if request.alias and request.alias in self._aliases.values():
                alias = request.alias

            if not alias and request.metric_name in self._datatypes:
                name = request.metric_name

            metric_ids.append(alias or name)
            values.append(request.value)

            if not alias and not name:
                not_found_metrics.append(index)
                continue

            datatype = self._datatypes.get(alias) or self._datatypes[name]
            metrics.append(Metric(timestamp, request.value, datatype, alias, name))

        if metrics:
            await self._mqtt_client.publish(
                SpbMessage(
                    Topic.from_component(self._node_name, "NCMD"),
                    NodePayload(timestamp, metrics),
                ),
                qos=0,
                retain=False,
            )

        track = _CommandTrack(metric_ids, values)
        if not_found_metrics:
            for index in not_found_metrics:
                track.resolutions[index] = WritingResolution.MetricNotFound

        if WritingResolution.WritingTimeout not in track.resolutions:
            track.has_finished.set()

        self._tracks[timestamp] = track



import abc
from dataclasses import dataclass
from typing import Any, Self

from ..data import Metric


@dataclass
class Topic:
    value: str

    @property
    def component_name(self) -> str:
        items = self.value.split("/")
        if items[1] == "STATE":
            return "/".join(items[2:])
        else:
            return "/".join([items[1]] + items[3:])

    @property
    def message_type(self) -> str:
        items = self.value.split("/")
        if items[1] == "STATE":
            return "STATE"
        return items[2]

    @classmethod
    def from_component(cls, name: str, message_type: str) -> Self:
        if message_type not in (
            "STATE",
            "NDATA",
            "NBIRTH",
            "NDEATH",
            "NCMD",
            "DDATA",
            "DBIRTH",
            "DDEATH",
        ):
            raise ValueError(f'Message type "{message_type}" is not a standard')
        if message_type == "STATE":
            return cls(f"spBv1.0/STATE/{name}")
        items = name.split("/")
        return cls(f"spBv1.0/{items[0]}/{message_type}/{'/'.join(items[1:])}")


@dataclass
class Payload(abc.ABC):
    timestamp: int

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, payload_map: dict[str, Any]) -> Self:
        """Create a payload from a plain dict"""

    @abc.abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Convert to dict for parsing"""


@dataclass
class SpbMessage:
    topic: Topic
    payload: Payload

    def is_a(self, message_type: str) -> bool:
        return message_type == self.topic.message_type

    def is_from_node(self) -> bool:
        return "STATE" != self.topic.message_type


@dataclass
class Will:
    """Will message to sent to MQTT broker"""

    message: SpbMessage
    qos: int
    retain: bool


class MqttClient(abc.ABC):
    @property
    @abc.abstractmethod
    def is_connected(self) -> bool:
        """Is the client connected to the Broker?"""

    @abc.abstractmethod
    async def connect(self, component_name: str, will: Will):
        """Connect a component to MQTT server"""

    @abc.abstractmethod
    async def publish(self, message: SpbMessage, qos: int, retain: bool):
        """Publish a message  to the topic"""

    @abc.abstractmethod
    async def deliver_message(self) -> SpbMessage:
        """Return a messsage recieved from the MQTT Server"""

    @abc.abstractmethod
    async def subscribe(self, topic: str, qos: int):
        """Subscribe the component to recieve messages from a topic"""

    @abc.abstractmethod
    async def disconnect(self):
        """Disconnect the client from the MQTT server"""


@dataclass
class HostPayload(Payload):
    """Payload to send by MQTT for State messages"""

    online: bool

    @classmethod
    def from_dict(cls, payload_map: dict[str, Any]) -> Self:
        """Create a payload from a plain dict"""
        return cls(payload_map["timestamp"], payload_map["online"])

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict for parsing"""
        return {"timestamp": self.timestamp, "online": self.online}


@dataclass
class MessageContent(abc.ABC):
    """Interface of a message content"""

    @abc.abstractmethod
    async def send(
        self,
        mqtt_client: "MqttClient",
        timestamp: int,
        component_name: str,
        seq: int | None = None,
    ):
        """Publish to MQTT Broker the content of the message"""


@dataclass
class NodePayload(Payload):
    """Payload to send by MQTT for/to edge node messages"""

    metrics: list[Metric]
    seq: int | None = None

    @classmethod
    def from_dict(cls, payload_map: dict[str, Any]) -> Self:
        """Create a payload from a plain dict"""
        return cls(
            payload_map["timestamp"],
            [Metric.from_dict(value) for value in payload_map["metrics"]],
            payload_map.get("seq"),
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict for parsing"""
        outcome = {
            "timestamp": self.timestamp,
            "metrics": [dto.as_dict() for dto in self.metrics],
        }
        if self.seq is not None:
            outcome["seq"] = self.seq

        return outcome


@dataclass
class MqttConfig:
    hostname: str
    port: int
    username: str | None = None
    password: str | None = None
    ca_certs: str = ""
    keepalive: int = 30


class MessageEncoder(abc.ABC):
    """Encode the message before sending it as payload in the message"""

    @abc.abstractmethod
    def encode(self, payload: Payload) -> bytes:
        """Convert a message to a payload"""

    @abc.abstractmethod
    def decode(self, payload: bytes) -> Payload:
        """Convert payload to a message object"""


class MqttError(Exception):
    """Wraps error of comunications by any MQTT adapter"""
